/*
 * kgraft_patch_bsc1156331
 *
 * Fix for CVE-2018-20856, bsc#1156331
 *
 *  Upstream commit:
 *  54648cf1ec2d ("block: blk_init_allocated_queue() set q->fq as NULL in the
 *                 fail case")
 *
 *  SLE12-SP1 commit:
 *  not affected
 *
 *  SLE12-SP2 and -SP3 commit:
 *  438ecbfbdcf6e19c8c5aa6f31608427cf79451ee
 *
 *  SLE12-SP4, SLE15 and SLE15-SP1 commit:
 *  8769ece87bcf3f003c550ee319bfb327333f8e5c
 *
 *
 *  Copyright (c) 2019 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/backing-dev.h>
#include <linux/bio.h>
#include <linux/blkdev.h>

#include <linux/kernel.h>
#include <linux/module.h>
#include "kgr_patch_bsc1156331.h"
#include "../kallsyms_relocs.h"


static struct blk_flush_queue *(*kgre_blk_alloc_flush_queue)(struct request_queue *q,
		int node, int cmd_size);
static void (*kgre_blk_free_flush_queue)(struct blk_flush_queue *q);

static int (*kgre_blk_init_rl)(struct request_list *rl, struct request_queue *q,
		gfp_t gfp_mask);

static void (*kgre_blk_timeout_work)(struct work_struct *work);


#ifdef CONFIG_BLK_WBT

static void (*kgre_wbt_exit)(struct request_queue *);

#else
#error "klp-ccp: non-taken branch"
#endif /* CONFIG_BLK_WBT */


static blk_qc_t (*kgre_blk_queue_bio)(struct request_queue *q, struct bio *bio);


/* from block/blk-core.c */
/* patched */
int kgrp_blk_init_allocated_queue(struct request_queue *q)
{
	q->fq = (*kgre_blk_alloc_flush_queue)(q, NUMA_NO_NODE, 0);
	if (!q->fq)
		return -ENOMEM;

	if ((*kgre_blk_init_rl)(&q->root_rl, q, GFP_KERNEL))
		goto fail;

	INIT_WORK(&q->timeout_work, (*kgre_blk_timeout_work));
	q->queue_flags		|= QUEUE_FLAG_DEFAULT;

	/*
	 * This also sets hw/phys segments, boundary and size
	 */
	blk_queue_make_request(q, (*kgre_blk_queue_bio));

	q->sg_reserved_size = INT_MAX;

	/* Protect q->elevator from elevator_change */
	mutex_lock(&q->sysfs_lock);

	/* init elevator */
	if (elevator_init(q, NULL)) {
		mutex_unlock(&q->sysfs_lock);
		goto fail;
	}

	mutex_unlock(&q->sysfs_lock);
	return 0;

fail:
	(*kgre_blk_free_flush_queue)(q->fq);
	/*
	 * Fix CVE-2018-20856
	 *  +1 line
	 */
	q->fq = NULL;
	(*kgre_wbt_exit)(q);
	return -ENOMEM;
}



static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "blk_alloc_flush_queue", (void *)&kgre_blk_alloc_flush_queue },
	{ "blk_free_flush_queue", (void *)&kgre_blk_free_flush_queue },
	{ "blk_init_rl", (void *)&kgre_blk_init_rl },
	{ "blk_timeout_work", (void *)&kgre_blk_timeout_work },
	{ "wbt_exit", (void *)&kgre_wbt_exit },
	{ "blk_queue_bio", (void *)&kgre_blk_queue_bio },
};

int kgr_patch_bsc1156331_init(void)
{
	return __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
}
